#include "common/utils.hpp"
#include "nodes/base/base.hpp"
#include "nodes/draw/drawNode.hpp"
#include "nodes/draw/position.hpp"
#include <opencv2/opencv.hpp>
#include <chrono>
#include <tuple>

namespace GNode
{

const std::vector<std::pair<int, int>> coco_pairs = {
    {0, 1}, {0, 2}, {0, 11}, {0, 12}, {1, 3}, {2, 4},
    {5, 6}, {5, 7}, {7, 9}, {6, 8}, {8, 10},
    {11, 12}, {5, 11}, {6, 12},
    {11, 13}, {13, 15}, {12, 14}, {14, 16}
};

static std::tuple<int, int, int> getFontSize(const std::string& text)
{
    int baseline = 0;
    cv::Size textSize = cv::getTextSize(text, cv::FONT_HERSHEY_SIMPLEX, 1.0, 2, &baseline);
    // std::cout << textSize << std::endl;
    return std::make_tuple(textSize.width, textSize.height, baseline);
}

static void overlay_mask(
    cv::Mat& image, const cv::Mat& smallMask,
    int roiX, int roiY,
    const cv::Scalar& color, double alpha) 
{
    if (image.empty() || smallMask.empty() ||
        image.type() != CV_8UC3 || smallMask.type() != CV_8UC1) {
        return;
    }
    alpha = std::max(0.0, std::min(1.0, alpha));
    
    cv::Rect roiRect(roiX, roiY, smallMask.cols, smallMask.rows);
    
    cv::Rect imageRect(0, 0, image.cols, image.rows);
    cv::Rect intersectionRect = roiRect & imageRect; // 使用 & 操作符计算交集
    
    if (intersectionRect.width <= 0 || intersectionRect.height <= 0) {
        return;
    }
    
    cv::Mat originalROI = image(intersectionRect); // ROI 指向 image 的数据
    
    int maskStartX = intersectionRect.x - roiX;
    int maskStartY = intersectionRect.y - roiY;
    cv::Rect maskIntersectionRect(maskStartX, maskStartY, intersectionRect.width, intersectionRect.height);
    cv::Mat smallMaskROI = smallMask(maskIntersectionRect);
    
    cv::Mat colorPatchROI(intersectionRect.size(), image.type(), color);
    
    cv::Mat tempColoredROI = originalROI.clone(); // 需要一个临时区域进行覆盖
    colorPatchROI.copyTo(tempColoredROI, smallMaskROI);
    
    cv::addWeighted(originalROI, 1.0 - alpha, tempColoredROI, alpha, 0.0, originalROI);
}


static std::tuple<uint8_t, uint8_t, uint8_t> hsv2bgr(float h, float s, float v) 
{
    const int h_i = static_cast<int>(h * 6);
    const float f = h * 6 - h_i;
    const float p = v * (1 - s);
    const float q = v * (1 - f * s);
    const float t = v * (1 - (1 - f) * s);
    float r, g, b;
    switch (h_i) {
      case 0:
        r = v, g = t, b = p;
        break;
      case 1:
        r = q, g = v, b = p;
        break;
      case 2:
        r = p, g = v, b = t;
        break;
      case 3:
        r = p, g = q, b = v;
        break;
      case 4:
        r = t, g = p, b = v;
        break;
      case 5:
        r = v, g = p, b = q;
        break;
      default:
        r = 1, g = 1, b = 1;
        break;
    }
    return std::make_tuple(static_cast<uint8_t>(b * 255), static_cast<uint8_t>(g * 255),
                      static_cast<uint8_t>(r * 255));
  }

static std::tuple<uint8_t, uint8_t, uint8_t> random_color(int id) 
{
    float h_plane = ((((unsigned int)id << 2) ^ 0x937151) % 100) / 100.0f;
    float s_plane = ((((unsigned int)id << 3) ^ 0x315793) % 100) / 100.0f;
    return hsv2bgr(h_plane, s_plane, 1);
}


static float box_iou(const data::Box& box1, const data::Box& box2) 
{
    float inter_x1 = std::max(box1.left, box2.left);
    float inter_y1 = std::max(box1.top, box2.top);
    float inter_x2 = std::min(box1.right, box2.right);
    float inter_y2 = std::min(box1.bottom, box2.bottom);

    float inter_area = std::max(0.0f, inter_x2 - inter_x1) * std::max(0.0f, inter_y2 - inter_y1);
    float box1_area = (box1.right - box1.left) * (box1.bottom - box1.top);
    float box2_area = (box2.right - box2.left) * (box2.bottom - box2.top);

    return inter_area / (box1_area + box2_area - inter_area + 1e-6f); // Avoid division by zero
}

void DrawNode::draw_track_boxes(cv::Mat& image, const data::BoxArray& track_boxes)
{
  auto now = std::chrono::steady_clock::now();

  for (const auto& box : track_boxes) 
  {
    uint8_t b, g, r;
    std::tie(b, g, r) = random_color(box.class_id);

    // Draw bounding box
    // cv::rectangle(image, cv::Point(box.left, box.top), cv::Point(box.right, box.bottom), cv::Scalar(b, g, r), 2);

    // Calculate center point of the box
    cv::Point center((box.left + box.right) / 2, (box.top + box.bottom) / 2);
    std::string track_text = str_format("ID: %d", box.class_id);
    cv::putText(image, track_text, center, cv::FONT_HERSHEY_SIMPLEX, 0.5, cv::Scalar(b, g, r), 2);
    track_history_[box.class_id].push_back(center);

    if (track_history_[box.class_id].size() > 50) 
    {
      track_history_[box.class_id].erase(track_history_[box.class_id].begin());
    }

    // Draw the track lines
    const auto& history = track_history_[box.class_id];
    for (size_t i = 1; i < history.size(); ++i) 
    {
      cv::line(image, history[i - 1], history[i], cv::Scalar(b, g, r), 2);
    }

    // Update the last seen time for this class ID
    last_track_update_time_[box.class_id] = now;
  }

  // Remove outdated tracks
  for (auto it = track_history_.begin(); it != track_history_.end(); ) 
  {
    if (last_track_update_time_.find(it->first) != last_track_update_time_.end() &&
      std::chrono::duration_cast<std::chrono::seconds>(now - last_track_update_time_[it->first]).count() > 5) 
    {
      last_track_update_time_.erase(it->first);
      it = track_history_.erase(it);
    } 
    else 
    {
      ++it;
    }
  }
}

void DrawNode::draw_boxes(cv::Mat& image, const data::BoxArray& boxes)
{
  int image_width = image.cols;
  int image_height = image.rows;
  PositionManager<float> pm(getFontSize);
  for (const auto& box : boxes) 
  {
    uint8_t b, g, r;
    std::tie(b, g, r) = random_color(box.class_id);
    cv::rectangle(image, cv::Point(box.left, box.top), cv::Point(box.right, box.bottom), cv::Scalar(b, g, r), 2);
    std::tuple<int, int, int, int> pbox = std::make_tuple(box.left, box.top, box.right, box.bottom);
    int x, y;
    std::string text = str_format("%s %.2f", box.label.c_str(), box.score);
    std::tie(x, y) = pm.selectOptimalPosition(pbox, image_width, image_height, text);
    
    cv::putText(image, text, cv::Point(x, y), cv::FONT_HERSHEY_SIMPLEX, 1, cv::Scalar(b, g, r), 2);
    if (!box.seg_mask.empty())
    {
        overlay_mask(image, box.seg_mask, box.left, box.top, cv::Scalar(b, g, r), 0.6);
    }
    if (!box.keypoints.empty())
    {
        for (const auto& point : box.keypoints)
        {
            cv::circle(image, cv::Point(point.x, point.y), 5, cv::Scalar(b, g, r), -1);
        }
        for (const auto& pair : coco_pairs) 
        {
            int startIdx = pair.first;
            int endIdx = pair.second;

            if (startIdx < box.keypoints.size() && endIdx < box.keypoints.size()) 
            {
                int x1 = (int)box.keypoints[startIdx].x;
                int y1 = (int)box.keypoints[startIdx].y;
                int x2 = (int)box.keypoints[endIdx].x;
                int y2 = (int)box.keypoints[endIdx].y;
                cv::line(image, cv::Point(x1, y1), cv::Point(x2, y2), cv::Scalar(b, g, r), 2);
            }
        }
    }
  }  
}

void DrawNode::draw_result(cv::Mat& image, const data::BoxArray& result)
{
  int image_width = image.cols;
  int image_height = image.rows;
  for (auto& box : result)
  {
      uint8_t b, g, r;
      std::tie(b, g, r) = random_color(box.class_id);
      cv::rectangle(image, cv::Point(box.left, box.top), cv::Point(box.right, box.bottom), cv::Scalar(b, g, r), 2);

      std::tuple<int, int, int, int> pbox = std::make_tuple(box.left, box.top, box.right, box.bottom);
      int x, y;
      std::string text = str_format("%s %.2f", box.label.c_str(), box.score);
      std::tie(x, y) = pm.selectOptimalPosition(pbox, image_width, image_height, text);
      
      cv::putText(image, text, cv::Point(x, y), cv::FONT_HERSHEY_SIMPLEX, 1.0f, cv::Scalar(b, g, r), 2);
  }
}


void DrawNode::handle_data(std::shared_ptr<meta::BaseData>& meta_data) 
{
    auto frame_data = std::dynamic_pointer_cast<meta::FrameData>(meta_data);
    if (config_data_ == nullptr) 
    {
        PLOGE.printf("DrawNode : [%s] config data is null, show nothing", name_.c_str());
        return;
    }
    if (config_data_->data_name != "DrawConfigData") 
    {
        PLOGE.printf("DrawNode : [%s] Invalid config data %s", name_.c_str(), config_data_->data_name.c_str());
        return;
    }
    auto config_data = std::dynamic_pointer_cast<meta::DrawConfigData>(config_data_);
    bool show_final_result = config_data->show_final_result;
    bool show_original_result = config_data->show_original_result;

    cv::Mat image = frame_data->image.clone();
    int image_width = image.cols;
    int image_height = image.rows;
    
    if (show_original_result) 
    {
        draw_boxes(image, frame_data->boxes);
        draw_track_boxes(image, frame_data->track_boxes);
    }

    if (show_final_result) 
    {
        draw_result(image, frame_data->result);
    }
    
    frame_data->draw_image = image;
}

}
